/**
 * 要使用本代码，页面需要加载 mpegts.js 和 hls.js 两个库
 */


/**
 * 播放器容器抽象，使用FlvPlayerWrapper/HlsNativePlayer/HlsJsPlayerWrapper等播放器时，要提供一个此抽象的实现类
 */
class PlayerContainer {


    /**
     * 播放器的ID，如一个页面内有多个时，用此ID区分。主要用于调试。
     *
     * @return {string}
     */
    id() {
    }

    /**
     * 是否正在请求打开流媒体或已经打开流媒体
     */
    isOpenRequested() {

    }

    /**
     * 当 PlayerWrapper 发生错误时，此方法被 PlayerWrapper 调用。实现类通常在此方法中在UI上提示用户。
     *
     * @param {string} errorMessage 错误信息
     */
    onPlayError(errorMessage) {
    }

    /**
     * 当开始播放时，此方法被 PlayerWrapper 调用。
     */
    onPlaying() {
    }

    /**
     * 当播放正常或异常结束时，此方法 PlayerWrapper 调用。实现类通常在此方法中做一些清理工作。
     */
    onClose() {
    }

}

/**
 * 播放器抽象
 */
class PlayerWrapper {

    /**
     *
     * @param {PlayerContainer} container PlayerContainer对象
     * @param {HTMLVideoElement} videoElmt 媒体播放元素
     * @param {string} mediaTyp 媒体类型：av - 音视频; a - 仅音频; v - 仅视频
     * @param {string} playUrl 媒体播放地址
     */
    constructor(container, videoElmt, mediaTyp, playUrl) {
        this.container = container;
        this.videoElmt = videoElmt;
        this.mediaTyp = mediaTyp;
        this.playUrl = playUrl;

        this.playing = false;
        this.playRequested = false;

        this.videoWidth = 0;
        this.videoHeight = 0;

        const that = this;

        this.videoElmt.addEventListener('loadedmetadata', function () {
            that.videoWidth = this.videoWidth;
            that.videoHeight = this.videoHeight;
        });
    }

    /**
     * 分辨率信息是否已经准备好
     *
     * @return {boolean}
     */
    get resolutionInfoReady() {
        return this.videoWidth !== 0;
    }

    /**
     * 播放器是否处于静音状态
     *
     * @return {boolean}
     */
    get muted() {
        return this.videoElmt.muted;
    }

    /**
     * 设置播放器静音
     *
     * @param value 是否要设置为静音
     */
    set muted(value) {
        this.videoElmt.muted = value;
    }

    /**
     * 加载媒体流并打开
     */
    load() {
    }

    /**
     * 暂停播放
     */
    pause() {
        this.videoElmt.pause();
        this.playRequested = true;
        this.playing = false;
    }

    /**
     * 停止播放
     */
    stop() {
        this.playing = false;
        this.playRequested = false;
    }

    _play() {
        this.playRequested = true;
    }

    _afterPlay() {
        this.playing = true;
        this.container.onPlaying();
    }

    _onPlayError(err) {
        this.container.onPlayError(err.toString());
        this.stop();
        this.container.onClose(); // TODO: is it needed?
        throw err;
    }

}

const flvJsSupported = mpegts.isSupported();


/**
 * FLV.js播放器封装
 */
class FlvPlayerWrapper extends PlayerWrapper {

    static checkSupported() {
        if (!flvJsSupported) {
            const error = '当前浏览器不支持 Media Source Extension，无法播放FLV格式的流媒体';
            console.error(error);
            throw new Error(error);
        }
    }

    /**
     *
     * @param {PlayerContainer} container
     * @param {HTMLVideoElement} videoElmt
     * @param {string} mediaTyp
     * @param {string} playUrl
     */
    constructor(container, videoElmt, mediaTyp, playUrl) {
        FlvPlayerWrapper.checkSupported();

        super(container, videoElmt, mediaTyp, playUrl);


        const mediaSource = {
            type: 'flv',
            isLive: true,
            url: playUrl
        };

        switch (mediaTyp) {
            case "av": // 音视频
                mediaSource.hasAudio = true;
                mediaSource.hasVideo = true;
                break;

            case "v": // 视频
                mediaSource.hasAudio = false;
                mediaSource.hasVideo = true;
                break;

            case "a": // 音频
                mediaSource.hasAudio = true;
                mediaSource.hasVideo = false;
                break;

            default:
                throw new Error(`Unsupported media type: ${mediaTyp}.`);
        }

        const config = {
            enableStashBuffer: false
        };


        /**
         *
         * @type {FlvPlayer}
         */
        this.player = mpegts.createPlayer(mediaSource);
        this.player.attachMediaElement(videoElmt);

        this.player.on(mpegts.Events.METADATA_ARRIVED, (_) => {
            if (!this.playRequested)
                this._play();
        });

        this.player.on(mpegts.Events.ERROR, (errType, detail) => {
            console.error(`player.onError(id=${this.container.id()}, errType=${errType}, detail=${detail})`);
            if (errType !== 'AbortError' && this.container.isOpenRequested()) {
                let message;
                if (errType === 'NetworkError' && detail === 'UnrecoverableEarlyEof' && this.playing)
                    message = '码流异常关闭';
                else
                    message = `${errType}: ${detail}`;
                this._onPlayError(message);
            }
        });
    }

    stop() {
        console.log('stop()');

        if (this.player != null) {
            this.player.unload();
            this.player.detachMediaElement();
            this.player.destroy();
            this.player = null;
        }

        super.stop();
    }

    load() {
        super.load();
        this.player.load();
    }

    /**
     *
     * @private
     */
    _play() {
        super._play();

        this.player.play()
            .then(() => {
                this._afterPlay();
            })
            .catch((err) => {
                let isAbort = false;
                if (err instanceof DOMException) {
                    if (err.name === 'AbortError')
                        isAbort = true;
                }

                if (!isAbort) {
                    this._onPlayError(err);
                }
            });

    }

}

const hlsJsSupported = Hls.isSupported();

/**
 * HLS.js播放器封装
 */
class HlsJsPlayerWrapper extends PlayerWrapper {

    static checkSupported() {
        if (!hlsJsSupported) {
            const error = '当前浏览器不支持 Media Source Extension，无法用 hls.js 进行流媒体播放';
            console.error(error);
            throw new Error(error);
        }
    }

    /**
     *
     * @param {PlayerContainer} container
     * @param {HTMLVideoElement} videoElmt
     * @param {string} mediaTyp
     * @param {string} playUrl
     */
    constructor(container, videoElmt, mediaTyp, playUrl) {
        HlsJsPlayerWrapper.checkSupported();

        super(container, videoElmt, mediaTyp, playUrl);

        const config = {
            debug: true
        };

        /**
         *
         * @type {Hls}
         */
        this.player = new Hls(config);
        this.player.attachMedia(videoElmt)

        this.player.on(Hls.Events.MANIFEST_PARSED, (_, data) => {
            console.log('manifest loaded, found ' + data.levels.length + ' quality level.');
            if (!this.playRequested)
                this._play();
        });

        this.player.on(Hls.Events.ERROR, (_, data) => {
            const errType = data.type;
            const errDetails = data.details;
            const errFatal = data.fatal;

            console.log(`Error occurred: ${JSON.stringify(data, null, 2)}.`);

            if (errFatal) {
                switch (errType) {
                    case Hls.ErrorTypes.NETWORK_ERROR:
                        console.log('fatal network error encountered, try to recover');
                        this.player.startLoad();
                        break;

                    case Hls.ErrorTypes.MEDIA_ERROR:
                        console.log('fatal media error encountered, try to recover');
                        this.player.recoverMediaError();
                        break;

                    default:
                        // cannot recover
                        this._onPlayError('播放时发生错误：' + errType + ', detail=' + errDetails);
                        this.stop();
                }

            }

        });
    }

    stop() {
        super.stop();
        this.player.destroy();
        this.player = null;
    }

    load() {
        super.load();
        this.player.loadSource(this.playUrl);
    }

    /**
     *
     * @private
     */
    _play() {
        super._play();

        this.videoElmt.play().then(() => {
            this._afterPlay();
        }).catch((err) => {
            this._onPlayError(err);
        });
    }
}

const hlsNativeSupport = MediaSource.isTypeSupported('application/vnd.apple.mpegurl');

const hlsSupported = hlsJsSupported || hlsNativeSupport;

/**
 * HLS原生播放器封装
 */
class HlsNativePlayer extends PlayerWrapper {

    static checkSupported() {
        if (!hlsNativeSupport) {
            const error = '当前浏览器不支持原生播放 HLS流媒体';
            console.error(error);
            throw new Error(error);
        }
    }

    /**
     *
     * @param {PlayerContainer} container
     * @param {HTMLVideoElement} videoElmt
     * @param {string} mediaTyp
     * @param {string} playUrl
     */
    constructor(container, videoElmt, mediaTyp, playUrl) {
        HlsNativePlayer.checkSupported();

        super(container, videoElmt, mediaTyp, playUrl);

        videoElmt.src = playUrl;
        videoElmt.onloadedmetadata = (event) => {
            if (!this.playRequested)
                this._play();
        };

        videoElmt.onerror = (event) => {
            this._onPlayError(`加载媒体时发生错误，媒体地址：${playUrl}, 错误信息：${JSON.stringify(event, null, 2)}`);
        };
    }


    stop() {
        super.stop();
        this.videoElmt.removeAttribute('src');
    }


    load() {
        this.videoElmt.load();
        super.load();
    }


    /**
     *
     * @private
     */
    _play() {
        super._play();

        this.videoElmt.play().then(() => {
            this._afterPlay();
        }).catch((err) => {
            this._onPlayError(err);
        });
    }
}
